#include "stdafx.h"





namespace ThinAudioGateway {





/*
################################################################################################

Implementation of DebugContext
================================================================================================

################################################################################################
*/





#if _Debug





DebugContextC::DebugContextC (DebugContext const * xParent, int const xPrintLevelOffset, PCTSTR const _Fmt, ...) {

  Assert (_Fmt != nullptr);

  SetParent (xParent);

  PrintLevelOffset = xPrintLevelOffset;

  va_list Args;

  va_start (Args, _Fmt);
  SetName (_Fmt, Args);
  va_end (Args);

}





void DebugContextC::SetParent (DebugContext const * const xParent) {

  Parent = xParent;
  Root = (Parent != nullptr)? ((Parent->Root != nullptr)? Parent->Root : Parent) : this;

  Assert (Root != nullptr);

}





void DebugContextC::SetName (PCTSTR const _Fmt, va_list const Args) {

  Assert (_Fmt != nullptr && Args != nullptr);

  wvsprintf (Name, _Fmt, Args);

  if (Parent != nullptr && Parent->Parent != nullptr) {

    wcscat (Name, L"/");
    wcscat (Name, Parent->Name);

  }

}





void _cdecl DebugContextC::SetName (PCTSTR const _Fmt, ...) {

  Assert (_Fmt != nullptr);

  va_list Args;

  va_start (Args, _Fmt);
  SetName (_Fmt, Args);
  va_end (Args);

}





void DebugContextC::SetPrintLevelOffset (int const Offset) {

  PrintLevelOffset = Offset;

}





//DebugPatch;
__pragma (check_stack (off))
__pragma (runtime_checks ("", off))
__pragma (optimize ("tg", on))





void DebugContextC::Printf (PCTSTR const _Fmt, va_list const Args, wchar_t const Delim) const {

  Assert (_Fmt != nullptr && Args != nullptr);

  wchar_t Buf [300];

  LARGE_INTEGER PF, PC;

  Verify (QueryPerformanceFrequency (&PF));
  Verify (QueryPerformanceCounter (&PC));

  /*
  If QPC time is wider than 44 bits, multiplying by 20-bit McsPerSecond causes an
  overflow. Since we need a 10-second remainder (0..9999999 mcs), it is enough to
  maintain a 34-bit counter at 10 GHz. So we can safely cut QPC value to 34 bits
  before converting it to microseconds.
  */

  UINT const Mcs = static_cast <UINT> (
    (static_cast <UINT64> (TestBits (PC.QuadPart, 0x3ffffffffUL)) * 1000000u / static_cast <UINT64> (PF.QuadPart))
  );

  int N = wsprintf (
    Buf,
    L"%s%c%07u%c%4u%c",
    (Root != nullptr)? Root->Name : Name, Delim, Mcs % 10000000, Delim, GetCurrentThreadId (), Delim
  );

  if (Parent != nullptr) {

    wcscat (Buf, Name);
    wcscat (Buf, L": ");
    N += static_cast <int> (wcslen (Name)) + 2;

  }

  N += wvsprintf (Buf + N, _Fmt, Args);

  wcscpy (Buf + N, L"\n");

  Assert (wcslen (Buf) < NumArrayElems (Buf));

  OutputDebugString (Buf);

}





void DebugContextC::Printf (UINT const Level, PCTSTR const _Fmt, va_list const Args) const {

  Assert (_Fmt != nullptr && Args != nullptr);

  UINT const ActualLevel = static_cast <UINT> (
    max (0, min (static_cast <int> (Level) + GetPrintLevelOffset (), 9))
  );

  if (ActualLevel >= GetBasePrintLevel ()) {

    wchar_t const Delim = (Level == 9)? L'*' : L' ';

    Printf (_Fmt, Args, Delim);

  }

}





void DebugContextC::Printf (UINT const Level, PCTSTR const _Fmt, ...) const {

  UINT const ActualLevel = static_cast <UINT> (
    max (0, min (static_cast <int> (Level) + GetPrintLevelOffset (), 9))
  );

  if (ActualLevel >= GetBasePrintLevel ()) {

    va_list Args;

    va_start (Args, _Fmt);
    Printf (Level, _Fmt, Args);
    va_end (Args);

  }

}





__pragma (check_stack ())
__pragma (runtime_checks ("", restore))
__pragma (optimize ("", on))





#else // _Debug

void DummyDebugFunc () { } // To avoid librarian warning 4221

#endif // _Debug





} // namespace
