namespace ThinAudioGateway {





typedef LocSize CopyDataFunc (void * const Context, void * const Addr, LocSize const Bytes);

PushWarnings ()

#if _MSC_VER >= 1800
  DisableWarnings (5204) // class has virtual functions, but its trivial destructor is not virtual
#endif

class CallbackWrapper : public CallbackInterface {

  FakeCopyConstructor (CallbackWrapper)
  FakeAssignmentOperator (CallbackWrapper)

  CopyDataFunc * const Func;
  void * const Context;

  LocSize CopyData (void * const Addr, LocSize const Bytes) {
    Assert (Func != nullptr);
    return Func (Context, Addr, Bytes);
  }

public:

  CallbackWrapper (CopyDataFunc * const xFunc, void * const xContext) : Func (xFunc), Context (xContext) { }

};

PopWarnings ()





class _LinkageDef TagHostC : public TagHost {

  FakeCopyConstructor (TagHostC)
  FakeAssignmentOperator (TagHostC)

  TagDriver const * const Driver;
  UINT const LineId;
  bool const Capture;
  CallbackInterface * const Callback;
  CallbackWrapper CbWrapper;

  HANDLE Event;
  TagPipe * Pipe;
  WaveFormat Fmt;
  bool Started;

  bool EventLoopActive;
  bool ExitRequest;

#if _Debug

  DebugContextC Dbg;

#endif // _Debug

public:

  TagHostC (TagDriver const * const Drv, UINT const LineId, bool const Capt, CallbackInterface * const CB);
  TagHostC (TagDriver const * const Drv, UINT const LineId, bool const Capt, CopyDataFunc * const CopyData, void * const Context);
  virtual ~TagHostC ();

  void PreInit ();

  UINT GetLineId () const { return LineId; }
  bool IsCapt () const { return Capture; }
  bool IsStarted () const { return Started; }
  PCTSTR GetDir () const { return IsCapt ()? L"capture" : L"render"; }
  CommonDataDesc const & GetCommonData () const { Assert (Pipe != nullptr); return Pipe->GetCommonData (); }

  static HRESULT _stdcall Create (
    TagHost * & Obj, TagDriver const * const Drv, UINT const LineId, bool const Capture, CallbackInterface * const CB
  );

  static HRESULT _stdcall Create (
    TagHost * & Obj, TagDriver const * const Drv, UINT const LineId, bool const Capture,
    CopyDataFunc * const CopyData, void * const Context
  );

  void Delete ();

  HRESULT Start (FormatRangeDesc const * const FormatRange, WaveFormatDesc const * const DefFormat, UINT const MinBufMcs);

  HRESULT Stop ();

  HRESULT EventLoop (UINT const PollPeriodMs);

#if _Debug

  void SetDebugPrintLevelOffset (int const Offset);

#endif // _Debug

  void Check () const {

#if _Debug

    Assert (Driver != nullptr);
    Assert (LineId != 0);
    Assert (!IsStarted () || Pipe != nullptr && Event != 0);

#endif

  }
  
};





} // namespace
