namespace ThinAudioGateway {





class _LinkageDef TagPipeC : public TagPipe {

  FakeAssignmentOperator (TagPipeC)
  FakeCopyConstructor (TagPipeC)

  TagDriver const * Drv;        // Driver reference
  UINT const LineId;            // Line Id
  bool const Capture;           // Capture flag

  UINT FrameSize;               // Frame size in bytes

  HANDLE Connection;            // Connection handle
  CommonDataDesc * CommonData;  // Driver/host common data block address

  HRESULT GetCommonDataAddress (CommonDataDesc * & CD) const;

#if _Debug

  DebugContextC Dbg;

#endif // _Debug

public:

  TagPipeC (TagDriver const * xDrv, UINT const xLineId, bool const xCapt);
  virtual ~TagPipeC ();

  UINT GetLineId () const { return LineId; }
  bool IsCapt () const { return Capture; }
  bool IsOpen () const { return (Connection != 0); }
  PCTSTR GetDir () const { return IsCapt ()? L"capture" : L"render"; }

  void Delete ();

  HRESULT Open ();
  void Close ();

  CommonDataDesc const & GetCommonData () const { Assert (CommonData != nullptr); return *CommonData; }
  CommonDataDesc & GetCommonData () { Assert (CommonData != nullptr); return *CommonData; }

  HRESULT ControlRequest (
    UINT const Func,
    DriverReq & Req, LocSize const ReqSize,
    void * const Data = nullptr, LocSize const DataSize = 0,
    LocSize * const RetSize = nullptr
  ) const;

  HRESULT SetFormatRange (FormatRangeDesc const & FRD) const;
  HRESULT SetDefaultFormat (WaveFormatDesc const & FD) const;

  HRESULT SetMinimumBufferLength (UINT const Mcs) const;
  HRESULT SetNotificationEvent (HANDLE const Event) const;

  HRESULT Notify (NTSTATUS const Code);

  HRESULT AdvanceDevicePosition (UINT const Frames);
  HRESULT ResetDevicePosition ();

#if _Debug

  void SetDebugPrintLevelOffset (int const Offset);

#endif // _Debug

  void Check () const {

#if _Debug

    Assert (Drv != nullptr);
    Assert (LineId != 0);
    Assert (Capture <= true);
    Assert (IsOpen () == (CommonData != nullptr));

#endif

  }
  
};





} // namespace
