#define _BuildMode_Lib 0
#define _BuildMode_Dll 1
#define _BuildMode_App 2



#define _Export _declspec (dllexport)
#define _Import _declspec (dllimport)



#if _BuildMode == _BuildMode_Lib

  __pragma (message ("Static library build mode"))

  #define _LinkageDecl _Export
  #define _LinkageDef _Export

#elif _BuildMode == _BuildMode_Dll
  
  __pragma (message ("DLL build mode"))

  #define _LinkageDecl
  #define _LinkageDef

#elif _BuildMode == _BuildMode_App
  
  __pragma (message ("Application build mode"))

  #define _LinkageDecl
  #define _LinkageDef

#else

  #error Incorrect _BuildMode value

#endif _BuildMode





namespace ThinAudioGateway {





/*
################################################################################################

Classes
================================================================================================

################################################################################################
*/





typedef UINT LocSize;





__if_not_exists (KSDATAFORMAT_WAVEFORMATEXTENSIBLE) {

  struct KSDATAFORMAT_WAVEFORMATEXTENSIBLE {

    KSDATAFORMAT DataFormat;
    WAVEFORMATEXTENSIBLE WaveFormatExt;

  };

}





#if _Debug

struct DebugContext {

  wchar_t Name [32];
  int PrintLevelOffset;         // Absolute level for the root desc, level offset for others

  DebugContext const * Parent;
  DebugContext const * Root;

};

#endif // _Debug





__interface CallbackInterface {

  LocSize CopyData (void * const Addr, LocSize const Bytes);

};





__interface TagPipe;





__interface TagDriver {

  bool IsFound () const;
  bool IsOpen () const;

  GUID const & GetClassId () const;
  PCTSTR GetInterface () const;

  void Delete ();

  HRESULT Find ();
  void Forget ();

  HRESULT Open ();
  void Close ();

  HRESULT ControlRequest (
    UINT const Func,
    DriverReq & Req, LocSize const ReqSize,
    void * const Data = nullptr, LocSize const DataSize = 0,
    LocSize * const RetSize = nullptr
  ) const;

  HRESULT GetInfo (DriverInfo & DI) const;
  HRESULT GetLineList (VirtualLineDesc * const Buffer, LocSize const Size, UINT & NumLines) const;
  HRESULT CreateLine (VirtualLineDesc const & Desc) const;
  HRESULT DeleteLine (UINT const Id) const;

  HRESULT CreatePipe (TagPipe * & Pipe, UINT const LineId, bool const Capture) const;

#if _Debug

  void SetDebugPrintLevelOffset (int const Offset);

#endif // _Debug

};





__interface TagPipe {

  bool IsCapt () const;
  bool IsOpen () const;
  PCTSTR GetDir () const;

  void Delete ();

  HRESULT Open ();
  void Close ();

  CommonDataDesc const & GetCommonData () const;
  CommonDataDesc & GetCommonData ();

  HRESULT ControlRequest (
    UINT const Func,
    DriverReq & Req, LocSize const ReqSize,
    void * const Data = nullptr, LocSize const DataSize = 0,
    LocSize * const RetSize = nullptr
  ) const;

  HRESULT SetFormatRange (FormatRangeDesc const & FRD) const;
  HRESULT SetDefaultFormat (WaveFormatDesc const & FD) const;

  HRESULT SetMinimumBufferLength (UINT const Mcs) const;
  HRESULT SetNotificationEvent (HANDLE const Event) const;

  HRESULT Notify (NTSTATUS const Code);

  HRESULT AdvanceDevicePosition (UINT const Frames);
  HRESULT ResetDevicePosition ();

#if _Debug

  void SetDebugPrintLevelOffset (int const Offset);

#endif // _Debug

};





__interface TagHost {

  UINT GetLineId () const;
  bool IsCapt () const;
  bool IsStarted () const;
  PCTSTR GetDir () const;
  CommonDataDesc const & GetCommonData () const;

  void Delete ();

  HRESULT Start (
    FormatRangeDesc const * const FormatRange, WaveFormatDesc const * const DefFormat, UINT const MinBufMcs
  );

  HRESULT Stop ();

  HRESULT EventLoop (UINT const PollPeriodMs);

#if _Debug

  void SetDebugPrintLevelOffset (int const Offset);

#endif // _Debug

};





/*
################################################################################################

Prototypes
================================================================================================

################################################################################################
*/





HRESULT _LinkageDecl _stdcall CreateDriver (TagDriver * & Obj, GUID const & ClassId);

HRESULT _LinkageDecl _stdcall CreateHost (
  TagHost * & Obj, TagDriver const * const Drv, UINT const LineId, bool const Capture, CallbackInterface * const CB
);





#if _Debug

  void SetDebugPrintLevel (UINT const Level);

#endif // _Debug





} // namespace
