namespace ThinAudioGateway {





template <UINT BufSize> class TextInPlace {

  public:

  // Buffer is defined public to allow derived classes to obtain its size via sizeof ().

  wchar_t Buf [BufSize];

  operator wchar_t const * () const { return Buf; }
  operator wchar_t * () { return Buf; }

};





class _LinkageDef ErrorText_i : public TextInPlace <256> {

  public:

  ErrorText_i (DWORD const Code) { ToStr (Buf, NumArrayElems (Buf), Code); }

  // HRESULT is LONG too

  ErrorText_i (LONG const Code) { ToStr (Buf, NumArrayElems (Buf), static_cast <DWORD> (Code)); }

  static LocSize ToStr (PTSTR const Buf, UINT const MaxLen, DWORD const Code);

};

// XxxText () functions must be macros to properly handle temporary objects

#define ErrorText(Code) (static_cast <PCTSTR> (ErrorText_i (Code)))
#define LastErrorText() (static_cast <PCTSTR> (ErrorText_i (GetLastError ())))





class GuidText_i : public TextInPlace <40> {

  public:

  GuidText_i (GUID const & Guid) { ToStr (Buf, NumArrayElems (Buf), Guid); }

  static LocSize ToStr (PTSTR const Buffer, UINT const MaxLen, GUID const & Guid);

};

// XxxText () functions must be macros to properly handle temporary objects

#define GuidText(Guid) (static_cast <PCTSTR> (GuidText_i (Guid)))





class _LinkageDef WaveFormat {

  WAVEFORMATEXTENSIBLE Fmt;

public:

  WaveFormat () { Init (); }

  WaveFormat (
    UINT const SR, UINT const BPS, UINT const NC, UINT const BPC = 0, UINT const CC = UINT_MAX
  ) {
    Init (SR, BPS, NC, BPC, CC);
  }

  WaveFormat (WaveFormatDesc const & FD) {
    Init (FD.SamplingRate, FD.BitsPerSample, FD.NumChannels, FD.BytesPerContainer, FD.ChannelConfig);
  }

  UINT GetTag () const { return Fmt.Format.wFormatTag; }
  UINT GetSubType () const;
  GUID const GetSubFormat () const;

  UINT GetSamplingRate () const { return Fmt.Format.nSamplesPerSec; }
  UINT GetBitsPerSample () const { return Fmt.Format.wBitsPerSample; }
  UINT GetActualBitsPerSample () const { return Fmt.Samples.wValidBitsPerSample; }
  UINT GetNumChannels () const { return Fmt.Format.nChannels; }
  UINT GetContainerSize () const { return GetBitsPerSample () / 8; }
  UINT GetChannelConfig () const { return Fmt.dwChannelMask; }
  UINT GetAlign () const { return Fmt.Format.nBlockAlign; }
  UINT GetAvgBytesPerSec () const { return Fmt.Format.nAvgBytesPerSec; }
  UINT GetFrameSize () const { return GetAlign (); }

  bool IsWfx () const; // Only WAVEFORMATEX part is valid
  bool IsPcm () const; // Describes a PCM format
  bool IsExtensible () const; // WAVEFORMATEXTENSIBLE descriptor is valid

  bool IsUniform () const { 
    return (GetAvgBytesPerSec () == GetSamplingRate () * GetFrameSize ());
  }

  // Does format descriptor contain format data (is filled)

  bool IsValid () const {
    return (GetTag () != WAVE_FORMAT_UNKNOWN || IsExtensible () && GetSubFormat () != GUID_NULL);
  }

  bool IsCorrect () const;

  operator WAVEFORMATEXTENSIBLE const & () const { return Fmt; }
  operator WAVEFORMATEX const & () const { return Fmt.Format; }

  WAVEFORMATEXTENSIBLE const & GetFormat () const { return Fmt; }
  WAVEFORMATEXTENSIBLE & GetFormat () { return Fmt; }

  WAVEFORMATEX const & GetWfx () const { return Fmt.Format; }
  WAVEFORMATEX & GetWfx () { return Fmt.Format; }

  static UINT GetStdChannelMask (UINT const NumChans);

  UINT BytesToFrames (UINT const Bytes) const { return Bytes / GetFrameSize (); }
  UINT FramesToBytes (UINT const Frames) const { return Frames * GetFrameSize (); }
  UINT MsToBytes (UINT const Ms) const { return FramesToBytes (Ms * Fmt.Format.nSamplesPerSec / 1000); }

  void Init () {
    ZeroObject (Fmt);
    Assert (!IsValid ());
  }

  HRESULT Init (
    UINT const SR, UINT const BPS, UINT const NC, UINT const BPC = 0, UINT const CC = UINT_MAX
  );

  HRESULT Init (WAVEFORMATEXTENSIBLE const & Wfx);

  void Invalidate () {
    Fmt.Format.wFormatTag = WAVE_FORMAT_UNKNOWN;
    Fmt.SubFormat = GUID_NULL;
  }

  static WaveFormat & FromWfx (WAVEFORMATEX & Wfx) {
    Assume (sizeof (WaveFormat) >= sizeof (Wfx));
    return reinterpret_cast <WaveFormat &> (Wfx);
  }

  static WaveFormat const & FromWfx (WAVEFORMATEX const & Wfx) {
    Assume (sizeof (WaveFormat) >= sizeof (Wfx));
    return const_cast <WaveFormat const &> (FromWfx (const_cast <WAVEFORMATEX &> (Wfx)));
  }

  static WaveFormat & FromWfx (WAVEFORMATEXTENSIBLE & Wfx) {
    Assume (sizeof (WaveFormat) == sizeof (Wfx));
    return reinterpret_cast <WaveFormat &> (Wfx);
  }

  static WaveFormat const & FromWfx (WAVEFORMATEXTENSIBLE const & Wfx) {
    Assume (sizeof (WaveFormat) == sizeof (Wfx));
    return const_cast <WaveFormat const &> (FromWfx (const_cast <WAVEFORMATEXTENSIBLE &> (Wfx)));
  }

  void ToWfx (WAVEFORMATEXTENSIBLE & Wfx) const { Wfx = Fmt; }
  void ToWfd (WaveFormatDesc & FD) const;

  LocSize ToStr (PTSTR const Buffer, UINT const MaxLen) const;

  void Check () const {

#if _Debug

    if (IsValid ()) {

      Assert (IsCorrect ());

    }

#endif _Debug

  }

};





class WfxText_i : public TextInPlace <50> {

  public:

  WfxText_i (WAVEFORMATEXTENSIBLE const & Wfx) {

    WaveFormat const & Fmt = Fmt.FromWfx (Wfx.Format);
    Fmt.ToStr (Buf, NumArrayElems (Buf));
    
  }

  WfxText_i (WAVEFORMATEX const & Wfx) {

    WaveFormat const & Fmt = Fmt.FromWfx (Wfx);
    Fmt.ToStr (Buf, NumArrayElems (Buf));
    
  }

  WfxText_i (WaveFormat const & Fmt) {

    Fmt.ToStr (Buf, NumArrayElems (Buf));
    
  }

  WfxText_i (WaveFormatDesc const & FD) {

    WaveFormat (FD).ToStr (Buf, NumArrayElems (Buf));
    
  }

};

// XxxText () functions must be macros to properly handle temporary objects

#define WfxText(Fmt) (static_cast <PCTSTR> (WfxText_i (Fmt)))





template <typename Type> class ScalarRange {

public:

  typedef ScalarRange <Type> RangeType;

private:

  Type Min, Max;

public:

  ScalarRange () : Min (0), Max (0) { }
  ScalarRange (Type const xMin, Type const xMax) : Min (xMin), Max (xMax) { }

  Type GetMin () const { return Min; }
  Type GetMax () const { return Max; }
  Type GetWidth () const { return Max - Min; }
  Type GetNumPoints () const { return Max - Min + 1; } // Is applicable only to discrete types

  void Get (Type & xMin, Type & xMax) const { xMin = Min; xMax = Max; }

  bool IsCorrect () const { return (Min <= Max); }

  bool operator == (ScalarRange const & R) const { return (Min == R.Min && Max == R.Max); }
  bool operator != (ScalarRange const & R) const { return !operator == (R); }

  bool IsValueWithin (Type const V) const {
    Assert (IsCorrect ());
    return (V >= Min && V <= Max);
  }

  bool IsRangeWithin (RangeType const & R) const {
    Assert (IsCorrect () && R.IsCorrect ());
    return (R.GetMin () >= Min && R.GetMax () <= Max);
  }

  bool IsWithinRange (RangeType const & R) const {
    Assert (IsCorrect () && R.IsCorrect ());
    return (Min >= R.GetMin () && Max <= R.GetMax ());
  }

  bool DoesIntersect (RangeType const & R) const {
    Assert (IsCorrect () && R.IsCorrect ());
    return (Min <= R.GetMax () && Max >= R.GetMin ());
  }

  // Must be defined inside to avoid the need of the explicit instantiation
  bool GetIntersection (RangeType const & R, RangeType & IS) const {
    Assert (IsCorrect () && R.IsCorrect ());
    if (DoesIntersect (R)) {
      IS.SetMin (max (Min, R.Min));
      IS.SetMax (min (Max, R.Max));
      Assert (IS.IsCorrect ());
      return true;
    } else {
      return false;
    }
  }

  Type LimitValue (Type const Value) const {

    return (max (min (Value, GetMax ()), GetMin ()));

  }

  // Must be defined here because depends on template parameters
  Type ScaleValueToRange (Type const Value, ScalarRange <Type> const & Target, bool const Round = false) const {

    Assert (GetWidth () > 0 && Target.GetWidth () > 0);

    return Round? (
      ((Value - GetMin ()) * Target.GetWidth () + GetWidth () / 2) / GetWidth () + Target.GetMin ()
    ) : (
      (Value - GetMin ()) * Target.GetWidth () / GetWidth () + Target.GetMin ()
    );

  }

  void SetMin (Type const V) { Min = V; }
  void SetMax (Type const V) { Max = V; }
  void Set (Type const xMin, Type const xMax) { SetMin (xMin); SetMax (xMax); }
  void Set (ScalarRange const & R) { Set (R.Min, R.Max); }

  // Must be defined here because depends on template parameters
  LocSize ToStr (PTSTR const Buffer, LocSize const MaxLen) const {

    int const IR = _snwprintf (Buffer, MaxLen, L"%u-%u", GetMin (), GetMax ());

    return (IR > 0)? static_cast <LocSize> (IR) : 0;

  }

  void Check () const {

#if _Debug

    Assert (IsCorrect ());

#endif _Debug

  }

};





//
// Convert a scalar range to text
// =============================================================================================
//

template <typename Type> class ScalarRangeText_i : public TextInPlace <24> {

public:

  typedef ScalarRange <Type> RangeType;

  ScalarRangeText_i (RangeType const & R) {

    R.ToStr (Buf, NumArrayElems (Buf));

  }

};

// XxxText () functions must be macros to properly handle temporary objects

#define ScalarRangeText(T,R) (static_cast <PCTSTR> (ScalarRangeText_i <T> (R)))





class _LinkageDef WaveFormatRange {

public:

  typedef UINT RangeBoundType;
  typedef ScalarRange <RangeBoundType> RangeType;

private:

  RangeType SamplingRate, BitsPerSample, NumChannels, BytesPerContainer;

public:

  WaveFormatRange () { }

  WaveFormatRange (RangeType const & SR, RangeType const & BPS, RangeType const & NC, RangeType const & BPC);

  WaveFormatRange (
    UINT const SrMin, UINT const SrMax,
    UINT const BpsMin, UINT const BpsMax,
    UINT const NcMin, UINT const NcMax,
    UINT const BpcMin, UINT const BpcMax
  ) {
    Set (SrMin, SrMax, BpsMin, BpsMax, NcMin, NcMax, BpcMin, BpcMax);
  }

  WaveFormatRange (FormatRangeDesc const & FRD) {
    Set (
      FRD.MinSamplingRate, FRD.MaxSamplingRate,
      FRD.MinBitsPerSample, FRD.MaxBitsPerSample,
      FRD.MinNumChannels, FRD.MaxNumChannels,
      FRD.MinBytesPerContainer, FRD.MaxBytesPerContainer
    );
  }

  bool operator == (WaveFormatRange const & R) const {
    return (
      SamplingRate == R.SamplingRate && BitsPerSample == R.BitsPerSample
      && NumChannels == R.NumChannels && BytesPerContainer == R.BytesPerContainer
    );
  }

  bool operator != (WaveFormatRange const & R) const { return !operator == (R); }

  RangeType const & GetSamplingRate () const { return SamplingRate; }
  RangeType const & GetBitsPerSample () const { return BitsPerSample; }
  RangeType const & GetNumChannels () const { return NumChannels; }
  RangeType const & GetBytesPerContainer () const { return BytesPerContainer; }

  bool IsCorrect () const;

  void SetSamplingRate (RangeType const & xSamplingRate) { SamplingRate.Set (xSamplingRate); }
  void SetBitsPerSample (RangeType const & xBitsPerSample) { BitsPerSample.Set (xBitsPerSample); }
  void SetNumChannels (RangeType const & xNumChannels) { NumChannels.Set (xNumChannels); }
  void SetBytesPerContainer (RangeType const & xBytesPerContainer) { BytesPerContainer.Set (xBytesPerContainer); }

  void Set (RangeType const & SR, RangeType const & BPS, RangeType const & NC, RangeType const & BPC) {
    SamplingRate.Set (SR);
    BitsPerSample.Set (BPS);
    NumChannels.Set (NC);
    BytesPerContainer.Set (BPC);
  }

  void Set (
    UINT const SrMin, UINT const SrMax,
    UINT const BpsMin, UINT const BpsMax,
    UINT const NcMin, UINT const NcMax,
    UINT const BpcMin, UINT const BpcMax
  ) {
    SamplingRate.Set (SrMin, SrMax);
    BitsPerSample.Set (BpsMin, BpsMax);
    NumChannels.Set (NcMin, NcMax);
    BytesPerContainer.Set (BpcMin, BpcMax);
  }

  LocSize ToStr (PTSTR const Buffer, LocSize const MaxLen) const;

  void ToFRD (FormatRangeDesc & FRD) {
    SamplingRate.Get (FRD.MinSamplingRate, FRD.MaxSamplingRate);
    BitsPerSample.Get (FRD.MinBitsPerSample, FRD.MaxBitsPerSample);
    NumChannels.Get (FRD.MinNumChannels, FRD.MaxNumChannels);
    BytesPerContainer.Get (FRD.MinBytesPerContainer, FRD.MaxBytesPerContainer);
  }

  void Check () const {

#if _Debug

    SamplingRate.Check ();
    BitsPerSample.Check ();
    NumChannels.Check ();

#endif _Debug

  }

};





//
// Convert a format range to text
// =============================================================================================
//

class WfrText_i : public TextInPlace <128> {

public:

  WfrText_i (WaveFormatRange const & FR) {

    FR.ToStr (Buf, NumArrayElems (Buf));

  }

  WfrText_i (FormatRangeDesc const & FRD) {

    WaveFormatRange (FRD).ToStr (Buf, NumArrayElems (Buf));

  }

};

// XxxText () functions must be macros to properly handle temporary objects

#define WfrText(R) (static_cast <PCTSTR> (WfrText_i (R)))





HRESULT IoControlRequest (
  HANDLE const Connection, UINT const Func,
  DriverReq & Req, LocSize const ReqSize,
  void * const Data = nullptr, LocSize const DataSize = 0,
  LocSize * const RetSize = nullptr
);





} // namespace
