/*

API definitions

*/





// Set 4-byte alignment to have the same alignment in both 32- and 64-bit code

__pragma (pack (push, 4))





namespace ThinAudioGateway {





/*
################################################################################################

Macros
================================================================================================

################################################################################################
*/





/*
API version
================================================================================================
*/

static UINT const ApiVersion = 5;





// Registry value names

#if 0

#define RegVal_MinSamplingRate L"Minimum sampling rate"
#define RegVal_MaxSamplingRate L"Maximum sampling rate"

#define RegVal_MinBitsPerSample L"Minimum bits per sample"
#define RegVal_MaxBitsPerSample L"Maximum bits per sample"

#define RegVal_MinNumChannels L"Minimum number of channels"
#define RegVal_MaxNumChannels L"Maximum number of channels"

#define RegVal_MinBytesPerContainer L"Minimum bytes per container"
#define RegVal_MaxBytesPerContainer L"Maximum bytes per container"

#define RegVal_DefSamplingRate L"Default sampling rate"
#define RegVal_DefBitsPerSample L"Default bits per sample"
#define RegVal_DefNumChannels L"Default number of channels"
#define RegVal_DefBytesPerContainer L"Default bytes per container"
#define RegVal_DefChannelConfig L"Default channel config"

#endif





/*
################################################################################################

Constants
================================================================================================

################################################################################################
*/





enum IoCtlFuncs : UINT {

  IOCF_TagFuncStart = 0x800,

  IOCF_DriverFuncStart = IOCF_TagFuncStart,

  IOCF_Driver_GetInfo,
  IOCF_Driver_GetLineList,
  IOCF_Driver_CreateLine,
  IOCF_Driver_DeleteLine,

  IOCF_DriverFuncEnd,

  IOCF_LineFuncStart = 0x900,

  IOCF_Line_GetCommonDataAddress,
  IOCF_Line_SetFormatRange,
  IOCF_Line_SetDefaultFormat,
  IOCF_Line_SetMinimumBufferLength,
  IOCF_Line_SetNotificationEvent,
  IOCF_Line_Notify,
  IOCF_Line_SetVolumeControlParams,

  IOCF_LineFuncEnd,

  IOCF_TagFuncEnd = IOCF_LineFuncEnd

};





#define DefineIoctlCode(Target, Name, Access) \
  static UINT const IoCtl_##Target##_##Name = CTL_CODE ( \
  FILE_DEVICE_UNKNOWN, \
  static_cast <UINT> (IOCF_##Target##_##Name), \
  METHOD_BUFFERED, \
  Access \
) \



DefineIoctlCode (Driver, GetInfo, FILE_READ_DATA);
DefineIoctlCode (Driver, GetLineList, FILE_READ_DATA);
DefineIoctlCode (Driver, CreateLine, FILE_WRITE_DATA);
DefineIoctlCode (Driver, DeleteLine, FILE_WRITE_DATA);

DefineIoctlCode (Line, GetCommonDataAddress, FILE_READ_DATA);
DefineIoctlCode (Line, Notify, FILE_WRITE_DATA);
DefineIoctlCode (Line, SetFormatRange, FILE_WRITE_DATA);
DefineIoctlCode (Line, SetDefaultFormat, FILE_WRITE_DATA);
DefineIoctlCode (Line, SetMinimumBufferLength, FILE_WRITE_DATA);
DefineIoctlCode (Line, SetNotificationEvent, FILE_WRITE_DATA);
DefineIoctlCode (Line, SetVolumeControlParams, FILE_WRITE_DATA);





#undef DefineLocalIoctlCode





static UINT const VirtualLineKsNameLenMax = 64; // Maximum length of KS virtual line friendly name
static UINT const VirtualLineEpNameLenMax = 32; // Maximum length of virtual line endpoint name





enum HostNotifications : UINT {

  HN_None,

  HN_StreamCreation,
  HN_StreamDeletion,
  HN_StreamStateChange,
  HN_VolumeChange,

};

#if _Debug

#define HostNotificationsEventNames \
  L"None", L"StreamCreation", L"StreamDeletion", L"StreamStateChange", L"VolumeChange"

#endif





/*
################################################################################################

Classes
================================================================================================

################################################################################################
*/





/*
Format range descriptor
================================================================================================
*/

struct FormatRangeDesc {

  UINT MinSamplingRate;         // Minimum sampling rate
  UINT MaxSamplingRate;         // Maximum sampling rate
  UINT MinBitsPerSample;        // Minimum bits per sample
  UINT MaxBitsPerSample;        // Maximum bits per sample
  UINT MinNumChannels;          // Minimum number of channels
  UINT MaxNumChannels;          // Maximum number of channels
  UINT MinBytesPerContainer;    // Minimum bytes per container
  UINT MaxBytesPerContainer;    // Maximum bytes per container

};





/*
Audio format descriptor
================================================================================================
*/

struct WaveFormatDesc {

  UINT SamplingRate;         // Sampling rate
  UINT BitsPerSample;        // Bits per sample
  UINT NumChannels;          // Number of channels
  UINT BytesPerContainer;    // Bytes per container
  UINT ChannelConfig;        // Channel configuration

};





/*
Volume control data descriptor
================================================================================================
*/

struct VolumeControlParamsDesc {

  bool Enabled;                 // Volume control enabled
  int MinDb;                    // Min level (dB)
  int MaxDb;                    // Max level (dB)
  UINT StepDb;                  // Step (dB)

};





/*
Volume levels descriptor
================================================================================================
*/

struct VolumeControlStateDesc {

  bool ChannelMutes [32];       // Channel mute state
  int ChannelLevels [32];       // Channel levels (dB * 65536)

};





/*
Main API classes
================================================================================================
*/





/*
Basic driver request descriptor (header)
================================================================================================
*/

struct DriverReq {

  UINT ApiVer;                  // API version

};





/*
Driver information request descriptor
================================================================================================
*/

struct DriverInfoReq : public DriverReq {

};





/*
Driver information descriptor
================================================================================================
*/

struct DriverInfo {

  UINT VerMajor;                // Major version number
  UINT VerMinor;                // Minor version number
  UINT VerRev;                  // Revision number
  UINT VerBuild;                // Build number
  FormatRangeDesc FormatRange;  // Driver's format range
  UINT NumLines;                // Number of virtual lines
  UINT Flags;                   // Driver flags

};





namespace DriverFlags {

}





/*
Virtual line (device) types
================================================================================================
*/

enum VirtualLineTypes : UINT {

  VLT_None,
  VLT_Microphone,
  VLT_LineConnector,
  VLT_Speaker,
  VLT_Headphones,
  VLT_SPDIF,

  VLT_Guard

};





/*
Virtual line (device) descriptor
================================================================================================
*/

struct VirtualLineDesc {

  UINT Id;                      // Line identifier
  bool Capture;                 // Capture/render line
  VirtualLineTypes Type;        // Line type (topology bridge pin category)
  wchar_t KsName [VirtualLineKsNameLenMax]; // KS friendly name (filter name)
  wchar_t EpName [VirtualLineEpNameLenMax]; // Endpoint name (in MediaCategories)

};





/*
Driver request descriptor: create virtual line/device
================================================================================================
*/

struct DriverReq_CreateLine : public DriverReq {

  VirtualLineDesc LineDesc;     // Line descriptor

};





/*
Driver request descriptor: delete virtual line/device
================================================================================================
*/

struct DriverReq_DeleteLine : public DriverReq {

  UINT Id;                      // Line identifier

};





__if_exists (WAVEFORMATPCMEX) {

/*
Filter information descriptor
================================================================================================
*/

struct FilterInfo {

  UINT Flags;                   // Filter flags
  HostNotifications NotificationCode; // Host notification code
  FormatRangeDesc FormatRange;  // Format range
  WaveFormatDesc DefaultFormat; // Default format
  UINT ReservedBufferSize;      // Reserved buffer size in bytes
  UINT MinBufferMcs;            // Minimum stream buffer duration in microseconds
  VolumeControlParamsDesc VolumeControlParams; // Volume control data
  bool HasStream;               // Has a stream
  UINT DataOverflows;           // Number of data overflows summed from all streams
  UINT DataUnderflows;          // Number of data underflows summed from all streams
  VolumeControlStateDesc VolumeState; // Current volume control state

};

} // if WAVEFORMATPCMEX is defined





namespace FilterFlags {

  UINT const Capture = 0x1;     // Capture filter
  UINT const StateChanged = 0x2; // Filter/stream state was changed since last notification
  UINT const WaitingForHost = 0x4; // Waiting for host confirmation
  UINT const All = 0x7;         // All flag bits

}





/*
Stream descriptor
================================================================================================
*/

__if_exists (WAVEFORMATEXTENSIBLE) {
__if_exists (KSSTATE) {

struct StreamInfo {

  UINT Flags;                   // Stream flags
  UINT ProcessId;               // Client process identifier
  UINT ThreadId;                // Client thread identifier
  WAVEFORMATEXTENSIBLE Fmt;     // Audio format
  KSSTATE State;                // Current state
  UINT BufferFrames;            // Buffer size in frames (0 - no buffer)
  UINT FramesBuffered;          // Amount of data buffered (frames)
  UINT64 FramesTransferred;     // Frames transferred
  UINT ClientPos;               // Client position in the buffer (frames)
  UINT DataOverflows;           // Number of data overflows
  UINT DataUnderflows;          // Number of data underflows

};

} // KSSTATE
} // WAVEFORMATPCMEX





namespace StreamFlags {

  UINT const Notifications = 1; // Notifications are used
  UINT const PacketMode = 2;    // Packet mode is used
  UINT const PositionRegister = 0x4; // Position register requested

  UINT const All = 0xf;         // All flag bits

}





//
// Two-way communication block for driver and host

#if _MSC_VER > 1500

  __pragma (warning (push))
  __pragma (warning (disable:
    4625 // copy constructor was implicitly defined as deleted
    4626 // assignment operator was implicitly defined as deleted
    5026 // move constructor was implicitly defined as deleted
    5027 // move assignment operator was implicitly defined as deleted
  ))

#endif

struct CommonDataDesc {

  struct {

    FilterInfo FI;              // Filter information block
    StreamInfo SI;              // Stream information block

  } Drv;                        // Driver-side data

  struct {

    void * Buffer;              // Client audio buffer mapped into host's space

#ifndef _WIN64

    UINT Reserved1;             // Add unused dword to align sizes in 32/64 environments

#endif

    UINT DevicePos;             // Host (device) position in the buffer (frames)

    UINT DevicePosBytes;        // Host (device) position in the buffer (bytes)
                                // If the position register is supported, this variable is mapped to the client's
                                // address space as a position register

    NTSTATUS Status;            // Status code from the host

    bool Ready;                 // Host side is ready

  } Host;                       // Host-side data

};

#if _MSC_VER > 1500

  __pragma (warning (pop))

#endif





/*
Special status codes
================================================================================================
*/

namespace StatusCodes {

  static NTSTATUS const CommonDataUpdated = 0x20000001;

}





/*
Basic filter request descriptor (header)
================================================================================================
*/

struct LineReq : public DriverReq {

};





/*
Request descriptor for getting common data address
================================================================================================
*/

struct LineReq_GetCommonDataAddress : public LineReq {

};





/*
Request descriptor for setting minimum buffer length/duration
================================================================================================
*/

struct LineReq_SetMinimumBufferLength : public LineReq {

  UINT Mcs;                     // Minimum microseconds

};





/*
Request descriptor for setting notification event
================================================================================================
*/

struct LineReq_SetNotificationEvent : public LineReq {

  HANDLE Event;                 // Event handle

};





/*
Request descriptor for driver notification
================================================================================================
*/

struct LineReq_NotifyDriver : public LineReq {

  NTSTATUS Code;                // Status code

};





/*
Request descriptor for setting format range
================================================================================================
*/

struct LineReq_SetFormatRange : public LineReq {

  FormatRangeDesc Range;        // Format range descriptor

};





/*
Request descriptor for setting default format
================================================================================================
*/

struct LineReq_SetDefaultFormat : public LineReq {

  WaveFormatDesc Format;       // Format descriptor

};





/*
Request descriptor for settting volume control parameters
================================================================================================
*/

struct LineReq_SetVolumeControlParams : public LineReq {

  VolumeControlParamsDesc Params; // Volume control parameters descriptor

};





/*
Request descriptor for setting volume state
================================================================================================
*/

struct LineReq_SetVolumeState : public LineReq {

  VolumeControlStateDesc VolumeControlState;  // Volume state descriptor

};





} // namespace





__pragma (pack (pop))
