#include "stdafx.h"





/*
################################################################################################

Constants
================================================================================================

################################################################################################
*/





static UINT const LineIdList [] = {

  1, 2, 3, 4, 5, 6, 7, 8, 9, UINT_MAX

};

static PCTSTR const LineDirList [] = {

  L"Render", L"Capture", nullptr

};

static PCTSTR const LineTypeList [] = {

  L"Microphone",
  L"Line connector",
  L"Speaker",
  L"Headphones",
  L"S/PDIF interface",

  nullptr

};





/*
################################################################################################

Implementation of Application class
================================================================================================

################################################################################################
*/





/*
Constructor
================================================================================================
*/

Application::Application (HINSTANCE const xInst)
:
  Inst (xInst),
  Window (0),
  StatusWindow (0),
  Driver (nullptr)

#if _Debug

  , Dbg (&::Dbg, 0, L"App")

#endif // _Debug

{

  ZeroObject (Workers);

}





/*
Destructor
================================================================================================
*/

Application::~Application () {

  for (UINT i = 0; i < NumArrayElems (Workers); ++i) {

    if (Workers [i] != nullptr) {

      Debug (6, L"Deleting worker %u", i);

      delete Workers [i];
      Workers [i] = nullptr;

    }

  }

  if (StatusWindow != 0) {

    Debug (6, L"Destroying status window %#x", StatusWindow);

    DestroyWindow (StatusWindow);
    StatusWindow = 0;

  }

  if (Window != 0) {

    Debug (6, L"Destroying main window %#x", Window);

    DestroyWindow (Window);
    Window = 0;

  }

  if (Driver != nullptr) {

    Debug (6, L"Deleting driver object");

#if UseLegacyWrappers

    Driver_Delete (Driver);

#else

    Driver->Delete ();

#endif

    Driver = nullptr;

  }

}





/*
Error message
================================================================================================
*/

void _cdecl Application::Error (PCTSTR const Format, ...) const {

  va_list Args;

  va_start (Args, Format);
  ::Error (Window, Format, Args);
  va_end (Args);

}





HRESULT Application::StartAll () {

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  //Assert (!Started);

  for (UINT i = 0; i < NumArrayElems (Workers); ++i) {

    if (Workers [i] != nullptr) {

      Debug (6, L"Starting worker %u", i);

      HR = Workers [i]->Start ();

      if (FAILED (HR)) {

        goto Exit;

      }

    }

  }

  //Started = true;

  Debug (7, L"Workers started");

  Assert (HR == S_OK);

Exit:

  if (FAILED (HR)) {

    StopAll ();

  }

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





HRESULT Application::StopAll () {

  for (UINT i = 0; i < NumArrayElems (Workers); ++i) {

    if (Workers [i] != nullptr) {

      Workers [i]->Stop ();

    }

  }

  //UpdateWindow (MainWin);

  //Started = false;

  return S_OK;

}





INT_PTR WINAPI Application::CreateLineDialogProc (HWND const Dlg, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  INT_PTR Res = 0;

  switch (Msg) {

    case WM_INITDIALOG: {

      Assert (Dlg != 0);

      SetWindowLongPtr (Dlg, DWLP_USER, LP);

      GblDebug (6, L"Dlg init, win %#x", Dlg);

      FillComboBox (Dlg, IDC_LineId, LineIdList);
      SendDlgItemMessage (Dlg, IDC_LineId, CB_SETCURSEL, 0, 0);

      FillComboBox (Dlg, IDC_LineDir, LineDirList);
      SendDlgItemMessage (Dlg, IDC_LineDir, CB_SETCURSEL, 0, 0);

      FillComboBox (Dlg, IDC_LineType, LineTypeList);
      SendDlgItemMessage (Dlg, IDC_LineType, CB_SETCURSEL, 0, 0);

      Assume (NumArrayElems (LineTypeList) == VLT_Guard);

      //ShowWindow (Dlg, SW_SHOW);

      break;

    }



    case WM_COMMAND: {

      UINT const CtrlId = LOWORD (WP);
      UINT const CtrlMsg = HIWORD (WP);

      Unreferenced (CtrlMsg);

      switch (CtrlId) {

        case IDOK: {

          Res = true;

          GblDebug (6, L"Ok");

          VirtualLineDesc * const VLD = reinterpret_cast <VirtualLineDesc *> (GetWindowLongPtr (Dlg, DWLP_USER));

          Assert (VLD != nullptr);

          RtlZeroMemory (VLD, sizeof (*VLD));

          BOOL IsTranslated;

          VLD->Id = GetDlgItemInt (Dlg, IDC_LineId, &IsTranslated, false);

          if (!IsTranslated || VLD->Id == 0) {

            ::Error (Dlg, L"Invalid line identifier");

            break;

          }

          LRESULT LR;
          
          LR = SendDlgItemMessage (Dlg, IDC_LineDir, CB_GETCURSEL, 0, 0);

          Assert (LR != CB_ERR && LR >= 0 && LR <= 1);

          VLD->Capture = (LR == 1);

          LR = SendDlgItemMessage (Dlg, IDC_LineType, CB_GETCURSEL, 0, 0);

          Assert (LR != CB_ERR && LR >= 0 && LR + 1 < VLT_Guard);

          VLD->Type = static_cast <VirtualLineTypes> (LR + 1);

          UINT N;
          
          N = GetDlgItemText (Dlg, IDC_LineNameKs, VLD->KsName, NumArrayElems (VLD->KsName));

          if (N == 0) {

            ::Error (Dlg, L"No KS filter/pin name");

            break;

          }

          N = GetDlgItemText (Dlg, IDC_LineNameEp, VLD->EpName, NumArrayElems (VLD->EpName));

          if (N != 0 && VLD->Type == VLT_Speaker) {

            ::Error (Dlg, L"Endpoint name doesn't make sense for speaker-type lines");

            break;

          }

          EndDialog (Dlg, S_OK);

          break;

        }

        case IDCANCEL: {

          Res = true;

          GblDebug (6, L"Cancel");

          EndDialog (Dlg, E_FAIL);

          break;

        }

      } // switch (CtrlId)

      break;

    } // WM_COMMAND

  }

  return Res;

}





INT_PTR WINAPI Application::DeleteLineDialogProc (HWND const Dlg, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  INT_PTR Res = 0;

  switch (Msg) {

    case WM_INITDIALOG: {

      Assert (Dlg != 0);

      SetWindowLongPtr (Dlg, DWLP_USER, LP);

      GblDebug (6, L"Dlg init, win %#x", Dlg);

      FillComboBox (Dlg, IDC_LineId, LineIdList);

      //ShowWindow (Dlg, SW_SHOW);

      break;

    }



    case WM_COMMAND: {

      UINT const CtrlId = LOWORD (WP);
      UINT const CtrlMsg = HIWORD (WP);

      Unreferenced (CtrlMsg);

      switch (CtrlId) {

        case IDOK: {

          Res = true;

          GblDebug (6, L"Ok");

          UINT * const LineId = reinterpret_cast <UINT *> (GetWindowLongPtr (Dlg, DWLP_USER));

          Assert (LineId != nullptr);

          BOOL IsTranslated;

          *LineId = GetDlgItemInt (Dlg, IDC_LineId, &IsTranslated, false);

          if (!IsTranslated || *LineId == 0) {

            ::Error (Dlg, L"Invalid line identifier");

            break;

          }

          EndDialog (Dlg, S_OK);

          break;

        }

        case IDCANCEL: {

          Res = true;

          GblDebug (6, L"Cancel");

          EndDialog (Dlg, E_FAIL);

          break;

        }

      } // switch (CtrlId)

      break;

    } // WM_COMMAND

  }

  return Res;

}





void Application::CreateLine () {

  HRESULT HR;
  
  VirtualLineDesc VLD;

  HR = static_cast <HRESULT> (DialogBoxParam (
    Inst, MAKEINTRESOURCE (IDD_CreateLine), GetWindow (), CreateLineDialogProc, reinterpret_cast <LPARAM> (&VLD)
  ));

  if (FAILED (HR)) {

    goto Exit;

  }

  Assert (VLD.Id != 0 && VLD.Type != VLT_None && VLD.Type < VLT_Guard && *VLD.KsName != L'\0');

  Assert (Driver != nullptr);

#if UseLegacyWrappers

  HR = Driver_CreateLine (Driver, &VLD);

#else

  HR = Driver->CreateLine (VLD);

#endif

  if (FAILED (HR)) {

    Error (L"Cannot create line (%s)", ErrorText (HR));

    goto Exit;

  }

  HR = CreateWorker (VLD);

  if (FAILED (HR)) {

    goto Exit;

  }

  Assert (HR == S_OK);

Exit:;

}





void Application::DeleteLine () {

  HRESULT HR;
  
  UINT LineId;

  HR = static_cast <HRESULT> (DialogBoxParam (
    Inst, MAKEINTRESOURCE (IDD_DeleteLine), GetWindow (), DeleteLineDialogProc, reinterpret_cast <LPARAM> (&LineId)
  ));

  if (FAILED (HR)) {

    goto Exit;

  }

  Assert (LineId != 0);

  Assert (Driver != nullptr);

#if UseLegacyWrappers

  HR = Driver_DeleteLine (Driver, LineId);

#else

  HR = Driver->DeleteLine (LineId);

#endif

  if (FAILED (HR)) {

    Error (L"Cannot delete line %u (%s)", LineId, ErrorText (HR));

    goto Exit;

  }

  HR = DeleteWorker (LineId);

  if (FAILED (HR)) {

    goto Exit;

  }

  Assert (HR == S_OK);

Exit:;

}





/*
Main window function
================================================================================================
*/

LRESULT Application::WindowProc (HWND const Win, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  Unreferenced (LP);

  LRESULT Res = -1;
  bool Handled = false;

  switch (Msg) {

    case WM_CREATE: {

      Handled = true;

      Debug (6, L"WM_CREATE, HWND %#x", Win);

      Assert (StatusWindow == 0);

      StatusWindow = CreateStatusWindow (
        WS_CHILD | WS_VISIBLE,
        L"Status",
        Win,
        StatusWindowId
      );

      if (StatusWindow == 0) {

        HRESULT const HR = HRESULT_FROM_WIN32 (GetLastError ());

        Error (L"Cannot create status window (%s)", ErrorText (HR));

        break;

      }

      Res = 0;

      break;

    }



    case WM_SIZE: {

      Handled = true;

      Assert (StatusWindow != 0);

      SendMessage (StatusWindow, WM_SIZE, WP, LP);

      Res = 0;

      break;

    }



    case WM_COMMAND: {

      UINT const CtrlId = LOWORD (WP);
      UINT const CtrlMsg = HIWORD (WP);

      Unreferenced (CtrlMsg);

      switch (CtrlId) {

        case ID_FILE_EXIT: {

          Debug (7, L"Exit command");

          StopAll ();
          PostQuitMessage (0);

          Handled = true;
          Res = 0;

          break;

        }

        case ID_LINE_CREATE: {

          Debug (6, L"Create line command");

          CreateLine ();

          Handled = true;
          Res = 0;

          break;

        }

        case ID_LINE_DELETE: {

          Debug (6, L"Delete line command");

          DeleteLine ();

          Handled = true;
          Res = 0;

          break;

        }

        case ID_WINDOW_TILE_H:
        case ID_WINDOW_TILE_V: {

          Debug (6, L"Tile windows command");

          TileWindows (
            GetWindow (),
            static_cast <UINT> ((CtrlId == ID_WINDOW_TILE_V)? MDITILE_VERTICAL : MDITILE_HORIZONTAL),
            nullptr, 0, nullptr
          );

          Handled = true;
          Res = 0;

          break;

        }

        case ID_WINDOW_CASCADE: {

          Debug (6, L"Cascade windows command");

          CascadeWindows (GetWindow (), 0, nullptr, 0, nullptr);

          Handled = true;
          Res = 0;

          break;

        }

      } // switch (CtrlId)
      
      break;

    } // WM_COMMAND



    case WM_CLOSE: {

      Debug (6, L"Close/exit");

      StopAll ();

      PostQuitMessage (0);
      Handled = true;
      Res = 0;

    }

  } // message switch

  if (!Handled) {

    Res = DefWindowProc (Win, Msg, WP, LP);

  }

  return Res;

}





/*
Main window function (outer)
================================================================================================
*/

LRESULT WINAPI Application::WindowProcOuter (HWND const Win, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  Application * App;

  if (Msg == WM_CREATE) {

    CREATESTRUCT const & CS = *reinterpret_cast <CREATESTRUCT const *> (LP);

    App = static_cast <Application *> (CS.lpCreateParams);

    SetWindowLongPtr (Win, GWLP_USERDATA, reinterpret_cast <LONG_PTR> (App));

  } else {

    App = reinterpret_cast <Application *> (GetWindowLongPtr (Win, GWLP_USERDATA));

  }

  return (App != nullptr)? App->WindowProc (Win, Msg, WP, LP) : DefWindowProc (Win, Msg, WP, LP);

}





/*
Main window creation
================================================================================================
*/

HRESULT Application::CreateMainWindow () {

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  Assert (Window == 0);

  {

    WNDCLASS WC;

    ZeroObject (WC);

    WC.lpszClassName = L"App window";
    WC.lpfnWndProc = WindowProcOuter;
    WC.hInstance = Inst;
    WC.hCursor = LoadCursor (nullptr, IDC_ARROW);
    WC.hbrBackground = reinterpret_cast <HBRUSH> (COLOR_WINDOWFRAME);

    ATOM const Class = RegisterClass (&WC);

    if (Class == 0) {

      HR = HRESULT_FROM_WIN32 (GetLastError ()); 

      Error (L"Cannot register main window class (%s)", ErrorText (HR));

      goto Exit;

    }

    HMENU const Menu = LoadMenu (Inst, MAKEINTRESOURCE (IDR_MainMenu));

    if (Menu == 0) {

      HR = HRESULT_FROM_WIN32 (GetLastError ());

      Error (L"Cannot load main menu (%s)", ErrorText (HR));

      goto Exit;

    }

    Debug (6, L"Creating main window");

    Window = CreateWindowEx (
      0 /*| WS_EX_CONTROLPARENT*/,
      reinterpret_cast <PCTSTR> (Class),
      L"Thin Audio Gateway test application",
      WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN /*| WS_HSCROLL | WS_VSCROLL*/,
      CW_USEDEFAULT, CW_USEDEFAULT, 900, 600,
      0, Menu, Inst, this
    );

    if (Window == 0) {

      HR = HRESULT_FROM_WIN32 (GetLastError ()); 

      Error (L"Cannot create main window (%s)", ErrorText (HR));

      goto Exit;

    }

    Debug (7, L"Created main window %#x", Window);

    HR = S_OK;

  }

Exit:

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





/*
Create worker for the line
================================================================================================
*/

HRESULT Application::CreateWorker (VirtualLineDesc const & VLD) {

  Worker * Wrk = nullptr;

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  {

    Debug (6, L"Creating worker for line %u", VLD.Id);

    UINT i;

    for (i = 0; i < NumArrayElems (Workers) && Workers [i] != nullptr; ++i);

    if (i == NumArrayElems (Workers)) {

      Error (L"No free worker entry");

      HR = HRESULT_FROM_WIN32 (ERROR_TOO_MANY_OPEN_FILES);
      goto Exit;

    }

    Wrk = new Worker (*this, VLD.Id, VLD.Capture, VLD.EpName);

    if (Wrk == nullptr) {

      Error (L"No memory for worker object");

      HR = E_OUTOFMEMORY;
      goto Exit;

    }

    HR = Wrk->Init ();

    if (FAILED (HR)) {

      goto Exit;

    }

    Workers [i] = Wrk;

    Debug (6, L"Worker for line %u added at index %u", Wrk->GetId (), i);

    Verify (SetWindowPos (
      *Wrk, 0,
      static_cast <int> (i * 50), static_cast <int> (i * 50), 0, 0,
      SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW
    ));

    Assert (HR == S_OK);

  }

Exit:

  if (FAILED (HR) && Wrk != nullptr) {

    delete Wrk;
    Wrk = nullptr;

  }

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





/*
Delete worker for the line
================================================================================================
*/

HRESULT Application::DeleteWorker (UINT const LineId) {

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  {

    Debug (6, L"Deleting worker for line %u", LineId);

    UINT i;

    for (i = 0; i < NumArrayElems (Workers) && (Workers [i] == nullptr || Workers [i]->GetId () != LineId); ++i);

    if (i == NumArrayElems (Workers)) {

      Error (L"No worker for line %u", LineId);

      HR = HRESULT_FROM_WIN32 (ERROR_NOT_FOUND);
      goto Exit;

    }

    Assert (Workers [i]->GetId () == LineId);

    Debug (6, L"Deleting worker for line %u at index %u", Workers [i]->GetId (), i);

    delete Workers [i];
    Workers [i] = nullptr;

    HR = S_OK;

  }

Exit:

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





/*
Run a message loop for main window
================================================================================================
*/

void Application::DoMessageLoop () const {

  Assert (Window != 0);

  Debug (6, L"Starting message loop");

  MSG Msg;

  while (GetMessage (&Msg, 0, 0, 0) > 0) {

    Debug (5, L"Msg loop: win %#x, msg %#x, WP %#x, LP %#x", Msg.hwnd, Msg.message, Msg.wParam, Msg.lParam);

    bool DialogMessage = false;

    for (UINT i = 0; i < NumArrayElems (Workers); ++i) {

      if (Workers [i] != nullptr && i == 1) {
        
        if (IsDialogMessage (*Workers [i], &Msg)) {

          DialogMessage = true;

          Debug (
            5,
            L"Msg %#x for worker %u dialog (win %#x)",
            Msg.message, Workers [i]->GetId (), Workers [i]->GetDlg ()
          );

          break;

        } else {

          Debug (
            5,
            L"Msg %#x is not for worker %u dialog (win %#x)",
            Msg.message, Workers [i]->GetId (), Workers [i]->GetDlg ()
          );

        }

      }

    } // Worker loop

    if (!DialogMessage) {

      TranslateMessage (&Msg);
      DispatchMessage (&Msg);

    }

  }

  Debug (6, L"Terminated message loop for window %#x", Window);

}





HRESULT Application::Run () {

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  Debug (6, L"Running");

  {

    InitCommonControls ();

    HR = CreateMainWindow ();

    if (FAILED (HR)) {

      goto Exit;

    }

    ShowWindow (Window, SW_SHOW);

    struct __declspec (uuid (ProductGuidStr)) ProductGuidStruct;

    GUID const & ProductGuid = __uuidof (ProductGuidStruct);

    Assert (Driver == nullptr);

#if UseLegacyWrappers

    HR = Driver_Create (reinterpret_cast <void **> (&Driver), &ProductGuid);

#else

    HR = ThinAudioGateway::CreateDriver (Driver, ProductGuid);

#endif

    if (FAILED (HR)) {

      Error (L"Cannot create driver object (%s)", ErrorText (HR));

      goto Exit;

    }



#if ConnectToDriver

#if UseLegacyWrappers

    HR = Driver_Find (Driver);

#else

    HR = Driver->Find ();

#endif

    if (FAILED (HR)) {

      if (HR == HRESULT_FROM_WIN32 (ERROR_NOT_FOUND)) {

        Error (L"No driver interface found");

      } else {

        Error (L"Cannot open the driver (%s)", ErrorText (HR));

      }

      goto Exit;

    }

#if UseLegacyWrappers

    HR = Driver_Open (Driver);

#else

    HR = Driver->Open ();

#endif

    if (FAILED (HR)) {

      Error (L"Cannot open the driver (%s)", ErrorText (HR));

      goto Exit;

    }

    DriverInfo DI;

    Driver->GetInfo (DI);

    VirtualLineDesc Lines [10];
    UINT NumLines;

    HR = Driver->GetLineList (Lines, sizeof (Lines), NumLines);

    if (FAILED (HR)) {

      Error (L"Cannot get line list from the driver (%s)", ErrorText (HR));

      goto Exit;

    }

#else

    static UINT const NumLines = 2;
    static DriverInfo const DI = {
      2, 0, 0, 0,
      {0, 0, 0, 0, 0, 0, 0, 0},
      NumLines,
      0
    };
    VirtualLineDesc const Lines [NumLines] = {
      {
        1,
        false,
        KSNODETYPE_SPEAKER,
        L"TAG Spk",
        L"Spk",
      },
      {
        2,
        true,
        KSNODETYPE_MICROPHONE,
        L"TAG Mic",
        L"Mic",
      },
    };

#endif

    wchar_t StatusText [120];

    wsprintf (
      StatusText,
      L"Driver version: %u.%u.%u.%u, format range: %s, number of lines: %u",
      DI.VerMajor, DI.VerMinor, DI.VerRev, DI.VerBuild, WfrText (DI.FormatRange), DI.NumLines
    );

    Assert (StatusWindow != 0);

    SetWindowText (StatusWindow, StatusText);

    for (UINT i = 0; i < NumLines; ++i) {

      VirtualLineDesc const & VLD = Lines [i];

      Debug (
        8,
        L"Line %u: %s, KS name: %s, EP name: %s",
        VLD.Id, VLD.Capture? L"capture" : L"render", VLD.KsName, VLD.EpName
      );

      HR = CreateWorker (VLD);

      if (FAILED (HR)) {

        goto Exit;

      }

    }

    DoMessageLoop ();

    HR = S_OK;

  }

Exit:

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}
