@echo off

setlocal

set CpuType=%~1
set Config=%~2

if "%CpuType%" == "" set CpuType=x86
if "%Config%" == "" set Config=release

if /i not "%CpuType%" == "x86" if /i not "%CpuType%" == "x64" if /i not "%CpuType%" == "arm64" (

  echo %~n0: Invalid CPU "%CpuType%"
  exit /b 1

)

call :Assert defined CppFiles
call :Assert defined OutFile



call "%~dp0setlocalenv"
if errorlevel 1 exit /b 1

call :Assert defined MsVcRootDir
call :Assert exist "%MsVcRootDir%\"

call :Assert defined WinSdkVer
call :Assert defined WinSdkRootDir
call :Assert exist "%WinSdkRootDir%\"



set MsVcBinDirTarget=%MsVcRootDir%\bin\Hostx86\%CpuType%
set MsVcIncDir=%MsVcRootDir%\include
set MsVcLibDir=%MsVcRootDir%\lib
set MsVcLibDirTarget=%MsVcRootDir%\lib\%CpuType%

call :Assert exist "%MsVcBinDirTarget%\cl.exe"
call :Assert exist "%MsVcBinDirTarget%\link.exe"

call :Assert exist "%MsVcIncDir%\stdarg.h"

call :Assert exist "%MsVcLibDirTarget%\libcmt.lib"



set WinSdkBinDirTarget=%WinSdkRootDir%\bin\%WinSdkVer%\x86

set WinSdkIncDir=%WinSdkRootDir%\Include\%WinSdkVer%
set WinSdkIncDirUm=%WinSdkIncDir%\um
set WinSdkIncDirShared=%WinSdkIncDir%\shared
set WinSdkIncDirCrt=%WinSdkIncDir%\ucrt

set WinSdkLibDir=%WinSdkRootDir%\Lib\%WinSdkVer%
set WinSdkLibDirUm=%WinSdkRootDir%\Lib\%WinSdkVer%\um\%CpuType%
set WinSdkLibDirCrt=%WinSdkRootDir%\Lib\%WinSdkVer%\ucrt\%CpuType%

call :Assert exist "%WinSdkBinDirTarget%\rc.exe"

call :Assert exist "%WinSdkIncDirUm%\windows.h"
call :Assert exist "%WinSdkIncDirShared%\devioctl.h"
call :Assert exist "%WinSdkIncDirCrt%\stddef.h"

call :Assert exist "%WinSdkLibDirUm%\kernel32.lib"
call :Assert exist "%WinSdkLibDirCrt%\ucrt.lib"



set path=%MsVcBinDirTarget%;%WinSdkBinDirTarget%;%path%
set include=%WinSdkIncDirUm%;%WinSdkIncDirShared%;%WinSdkIncDirCrt%;%MsVcIncDir%;%include%
set lib=%WinSdkLibDirUm%;%WinSdkLibDirCrt%;%MsVcLibDirTarget%;%lib%

::set path
::set include
::set lib



set OutDir=out\%Config%\%CpuType%

for %%F in (%OutFile%) do set OutFileBaseName=%%~nF
for %%F in (%OutFile%) do set OutFileExt=%%~xF

set PrepDefs=/DUNICODE /D_UNICODE /DFinalBuild=1 %PrepDefs%
set CppOpts=/nologo /c /Zi /MP /Fo"%OutDir%\\" /Fd"%OutDir%\\" %CppOpts%
set LinkOpts=/nologo /opt:ref /opt:icf /debug /pdbcompress /pdbaltpath:"%OutFileBaseName%.pdb" %LinkOpts%

if /i "%OutFileExt%" == ".dll" (

  set LinkOpts=/dll /implib:"%OutDir%\%OutFileBaseName%.lib" %LinkOpts%

)

if /i "%Config%" == "debug" (

  set PrepDefs=%PrepDefs% /DDEBUG
  set CppOpts=%CppOpts% /Od /MTd
  set CfgSfx=d

) else if /i "%Config%" == "release" (

  set PrepDefs=%PrepDefs% /DNDEBUG
  set CppOpts=%CppOpts% /O2 /MT
  set CfgSfx=

) else (

  echo %~n0: Invalid config "%Config%"
  exit /b 1

)



echo Building %OutFile% (%Config%, %CpuType%)

set ObjFiles=

for %%F in (%CppFiles%) do (

  call set ObjFiles=%%ObjFiles%% "%OutDir%\%%~nF.obj"

)

if not exist "%OutDir%\" (

  md "%OutDir%"
  if errorlevel 1 exit /b 1

)



echo Compiling C++ files

set RspFile="%OutDir%\cl.rsp"

echo %PrepDefs% %CppOpts% %CL_CustomVerDefs% %CppFiles% > "%RspFile%"

cl "@%RspFile%"
if errorlevel 1 exit /b 1

set ObjFiles=

for %%F in (%CppFiles%) do (

  call set ObjFiles=%%ObjFiles%% "%OutDir%\%%~nF.obj"

)



if defined RcFiles (

  echo Compiling resources

  set ResFiles=

  for %%F in (%RcFiles%) do (

    rc %PrepDefs% "/fo%OutDir%\%%~nF.res" "%%F"
    if errorlevel 1 exit /b 1

    call set ResFiles=%%ResFiles%% "%OutDir%\%%~nF.res"

  )

)



if /i "%OutFileExt%" == ".lib" (

  echo Creating library %OutFile%

  link /lib /nologo /out:"%OutDir%\%OutFile%" %ObjFiles%
  if errorlevel 1 exit /b 1

)



set UseLink=

if /i "%OutFileExt%" == ".exe" set UseLink=1
if /i "%OutFileExt%" == ".dll" set UseLink=1

if defined UseLink (

  echo Linking %OutFile%

  setlocal enabledelayedexpansion

  set CmdLine=%LinkOpts% /out:"%OutDir%\%OutFile%" %ObjFiles% %ResFiles% %LibFiles%

  set CmdLine=!CmdLine:$OutDir=%OutDir%!

  set RspFile=%OutDir%\link.rsp

  echo !CmdLine! > "!RspFile!"

  link "@!RspFile!"
  if errorlevel 1 exit /b 1

  setlocal disabledelayedexpansion

)



echo "%OutDir%\%OutFile%" successfully built

exit /b 0



:Assert

if %* (

  rem

) else (

  echo Assertion failed: %*

  rem Exit from all command files
  exit 1

)
