#include "common.h"






// ########################################################################
//
// Compiler and linker parameters
// =========
//
// ########################################################################





__pragma (comment (linker, "/subsystem:windows"))
__pragma (comment (linker, "/version:5.0"))

//__pragma (comment (lib, "kernel32.lib"))
__pragma (comment (lib, "user32.lib"))
__pragma (comment (lib, "advapi32.lib"))
__pragma (comment (lib, "winmm.lib"))
__pragma (comment (lib, "setupapi.lib"))
__pragma (comment (lib, "newdev.lib"))





// ########################################################################
//
// Constants
// =========
//
// ########################################################################





// ########################################################################
//
// Variables
// =========
//
// ########################################################################





OSVERSIONINFO VersionInfo;

bool NoInfoMessages = false;
bool NoErrorMessages = false;
bool NoFlush = false;
bool ProcessingStarted = false;

wchar_t InstancePath [MAX_PATH];

TOKEN_USER * UserToken;         // User token
TOKEN_OWNER * OwnerToken;       // Owner token





// ########################################################################
//
// Stub functions
// ==============
//
// ########################################################################





// Using no SSE and/or vectored memset/memcpy. These stubs prevent SSE2 code from
// linking from the LIBC.

extern "C" UINT __sse2_available = 0;

extern "C" void _cdecl _VEC_memzero () { Break (); }
extern "C" void _cdecl _VEC_memcpy () { Break (); }





// ########################################################################
//
// Functions
// =========
//
// ########################################################################





static bool Uninstall (PCTSTR const Id, bool & RebootRequired) {

  HDEVINFO DL;

  bool Success = false;

  do {

    DL = CreateDL (nullptr, DIGCF_ALLCLASSES | DIGCF_PRESENT);

    if (DL == 0) {

      break;

    }

    SP_DEVINFO_DATA DI;

    if (!FindDeviceById (DL, DI, Id)) {

      Error (L"Device for %s not found", Id);

      break;

    }

    wchar_t InstanceId [MAX_PATH];

    if (SetupDiGetDeviceInstanceId (DL, &DI, InstanceId, NumArrayElems (InstanceId), nullptr)) {

      Debug (7, L"Instance ID: \"%s\"", InstanceId);

    } else {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device instance ID (%s)", LastErrorText ());

      SetEOS (InstanceId);

    }

    wchar_t Desc [MAX_PATH];

    if (
      SetupDiGetDeviceRegistryProperty (
        DL,
        &DI,
        SPDRP_DEVICEDESC,
        nullptr,
        reinterpret_cast <PBYTE> (Desc),
        sizeof (Desc),
        nullptr
      )
    ) {

      Debug (6, L"Desc: \"%s\"", Desc);

    } else {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device desc (%s)", LastErrorText ());

      SetEOS (Desc);

    }

    wchar_t Service [MAX_PATH];

    if (
      SetupDiGetDeviceRegistryProperty (
        DL,
        &DI,
        SPDRP_SERVICE,
        nullptr,
        reinterpret_cast <PBYTE> (Service),
        sizeof (Service),
        nullptr
      )
    ) {

      Debug (6, L"Service ID: \"%s\"", Service);

    } else {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device service (%s)", LastErrorText ());

      SetEOS (Service);

    }

    wchar_t Driver [256];
    wchar_t InfPath [256];

    HKEY DrvKey = SetupDiOpenDevRegKey (
      DL,
      &DI,
      DICS_FLAG_GLOBAL,
      0,
      DIREG_DRV,
      GENERIC_READ
    );

    if (DrvKey != INVALID_HANDLE_VALUE) {

      if (GetRegistrySzValue (DrvKey, L"Driver", Driver, sizeof (Driver)))  {

        Debug (6, L"Driver: \"%s\"", Driver);

      } else {

        SetEOS (Driver);

      }

      if (GetRegistrySzValue (DrvKey, L"InfPath", InfPath, sizeof (InfPath)))  {

        Debug (6, L"InfPath: \"%s\"", InfPath);

      } else {

        SetEOS (InfPath);

      }

      RegCloseKey (DrvKey);
      DrvKey = 0;

    } else {

      DrvKey = 0;

      TranslateSetupApiError ();

      Warning (L"Cannot open driver key (%s)", LastErrorText ());

      SetEOS (Driver);
      SetEOS (InfPath);

    }



    StopDeviceDriverWithRetry (DL, DI, Desc, Service);



    bool TracesLeft = false;



    if (!RemoveDevice (DL, DI, InstanceId)) {

      break;

    }

    CheckIfRebootRequired (DL, DI, RebootRequired);

    if (!IsEOS (InfPath)) {

      Assert (wcschr (InfPath, L'\\') == nullptr);

      // SetupUninstallOEMInf removes driver package as well

      Trace (7, L"Calling SetupUninstallOEMInf (\"%s\", 0, nullptr)", InfPath);

      if (SetupUninstallOEMInf (InfPath, 0, nullptr)) {

        Trace (8, L"INF/PNF deleted for %s", InfPath);

      } else {

        TranslateSetupApiError ();

        Trace (9, L"Cannot uninstall %s (%s)", InfPath, LastErrorText ());

        if (GetLastError () != ERROR_FILE_NOT_FOUND && GetLastError () != ERROR_PATH_NOT_FOUND) {

          Warning (L"Cannot uninstall %s (%s)", InfPath, LastErrorText ());

          TracesLeft = true;

        }

      }

    }

    if (!IsEOS (Service)) {

      if (!RemoveService (Service) && GetLastError () != ERROR_SERVICE_DOES_NOT_EXIST) {

        TracesLeft = true;

      }

    }

    if (!IsEOS (Driver)) {

      wchar_t FullPath [256];

      GetWindowsDirectory (FullPath, NumArrayElems (FullPath));

      wcscat (FullPath, L"\\system32\\drivers\\");
      wcscat (FullPath, Driver);

      if (
        !DelFile (FullPath)
        && GetLastError () != ERROR_FILE_NOT_FOUND
        && GetLastError () != ERROR_PATH_NOT_FOUND
      ) {

        TracesLeft = true;

      }

    }

    if (RebootRequired && !TracesLeft) {

      // If there are no traces left, reboot is not required

      RebootRequired = false;

    }

    Success = true;

  } OnlyOnce;

  DestroyDL (DL);

  return Success;

}





static bool Install (PCTSTR const HardwareId, PCTSTR const InfFile, bool & RebootRequired, ULONG & DeviceProblem) {

  Assert (!IsEOS (HardwareId) && wcslen (HardwareId) < MAX_PATH);
  Assert (!IsEOS (InfFile) && wcslen (InfFile) < MAX_PATH);

  HDEVINFO DL = 0;
  SP_DEVINFO_DATA DI;

  ZeroObject (DI);

  Debug (7, L"Installing \"%s\" from \"%s\"", HardwareId, InfFile);

  bool DeviceInstanceCreated = false;
  wchar_t InstanceId [MAX_PATH];

  wcscpy (InstanceId, L"?");

  bool Success = false;

  do {

    for (UINT i = 0; i < 2; i++) {

      wchar_t Name [MAXPNAMELEN];

      if (GetPreferredDevice ((i != 0), Name)) {

        SetPreferredDevice ((i != 0), Name);

      }

    }

    GUID ClassGuid;
    wchar_t ClassName [MAX_CLASS_NAME_LEN];

    if (
      !SetupDiGetINFClass (
        InfFile,
        &ClassGuid,
        ClassName, NumArrayElems (ClassName),
        nullptr
      )
    ) {

      TranslateSetupApiError ();

      Error (L"Cannot get INF class (%s)", LastErrorText ());

      break;

    }

    Debug (7, L"INF class name is \"%s\"", ClassName);

    Assert (DL == 0);

    DL = SetupDiCreateDeviceInfoList (&ClassGuid, 0);

    if (DL == INVALID_HANDLE_VALUE) {

      DL = 0;

      TranslateSetupApiError ();

      Error (L"Cannot create device info list (%s)", LastErrorText ());

      break;

    }

    wchar_t InstanceDesc [MAX_PATH];

    wsprintf (
      InstanceDesc,
      L"%s%s",
      (!IsEOS (InstancePath))? InstancePath : ClassName,
      (!IsEOS (InstancePath))? L"" : L" (autogenerated)"
    );

    HRESULT HR = E_FAIL;
    bool SecondTry = false;

    for (;;) {

      Debug (7, L"Creating device instance \"%s\"", InstanceDesc);

      ZeroObject (DI);
      DI.cbSize = sizeof (DI);

      if (
        SetupDiCreateDeviceInfo (
          DL,
          !IsEOS (InstancePath)? InstancePath : ClassName,
          &ClassGuid,
          nullptr,
          0,
          static_cast <DWORD> ((!IsEOS (InstancePath))? 0 : DICD_GENERATE_ID),
          &DI
        )
      ) {

        HR = S_OK;
        break;

      }

      HR = TranslateSetupApiError ();

      Debug (9, L"Cannot create device instance \"%s\" (%s)", InstanceDesc, ErrorText (HR));

      if (HR != SPAPI_E_DEVINST_ALREADY_EXISTS || SecondTry || IsEOS (InstancePath)) {

        Assert (HR != S_OK);

        Error (L"Cannot create device instance \"%s\" (%s)", InstanceDesc, ErrorText (HR));

        break;

      }

      Assert (!IsEOS (InstancePath));

      Trace (8, L"Trying to delete the \"phantom\" mark for device instance \"%s\"", InstancePath);

      HR = DeletePhantomValue (InstancePath);

      if (FAILED (HR)) {

        SetLastError (static_cast <DWORD> (HR));
        break;

      }

      //DestroyDL (DL);

      if (!SetupDiOpenDeviceInfo (DL, InstancePath, 0, 0, &DI)) {

        HR = TranslateSetupApiError ();

        Error (L"Cannot open device instance \"%s\" (%s)", InstancePath, ErrorText (HR));

        break;

      }

      Debug (7, L"Device instance \"%s\" is opened", InstancePath);

      if (!RemoveDevice (DL, DI, InstancePath)) {

        break;

      }

      SecondTry = true;

    } // retry loop

    if (FAILED (HR)) {

      break;

    }

    DeviceInstanceCreated = true;

    Debug (8, L"Created device instance, node %#x", DI.DevInst);

    // Until the device is registered, it is marked by the nonzero Phantom property
    // value (REGSTR_VAL_PHANTOM) so Config Manager does not enumerate it.

#if 0 && _Debug

    __pragma (message ("##### Debug-only code in effect! #####"))
    //Debug (9, L"Pausing...");
    //Sleep (30 * 1000);
    //if (DeviceInstanceCreated) { ExitProcess (1); }

#endif

    if (!IsEOS (InstancePath)) {

      wcscpy (InstanceId, InstancePath);

    } else if (!SetupDiGetDeviceInstanceId (DL, &DI, InstanceId, NumArrayElems (InstanceId), nullptr)) {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device instance ID (%s)", LastErrorText ());

      wcscpy (InstanceId, L"?");

      break;

    }

    Debug (8, L"Instance ID for node %#x: \"%s\"", DI.DevInst, InstanceId);

    wchar_t HwId [MAX_PATH + 1];

    Assert (wcslen (HardwareId) + 1 < NumArrayElems (HwId));

    wcscpy (HwId, HardwareId);

    SetEOS (HwId + wcslen (HwId) + 1);

    Debug (7, L"Setting SPDRP_HARDWAREID to \"%s\"", HwId);

    if (
      !SetupDiSetDeviceRegistryProperty (
        DL, &DI,
        SPDRP_HARDWAREID,
        reinterpret_cast <PBYTE> (HwId), StrLenToSz (wcslen (HwId) + 2)
      )
    ) {

      TranslateSetupApiError ();

      Error (
        L"Cannot set SPDRP_HARDWAREID for \"%s\" to \"%s\" (%s)",
        InstanceId, HwId, LastErrorText ()
      );

      break;

    }



#if 0 && _Debug

    __pragma (message ("##### Debug-only code in effect! #####"))
    Debug (9, L"Pausing...");
    Sleep (30 * 1000);

#endif

#if 0 && _Debug

    __pragma (message ("##### The process exits after device creation! #####"))
    if (GetTickCount () != 0) ExitProcess (1111);

#endif



    // This call removes Phantom property value (REGSTR_VAL_PHANTOM) from the device data

    if (!SetupDiCallClassInstaller (DIF_REGISTERDEVICE, DL, &DI)) {

      TranslateSetupApiError ();

      Error (L"Cannot register device instance \"%s\" (%s)", InstanceId, LastErrorText ());

      break;

    }

    if (!EnableDevice (DL, DI, false)) {

      break;

    }

    Debug (
      7,
      L"Calling UpdateDriverForPlugAndPlayDevices (HwId=%s, Inf=\"%s\")",
      HardwareId, InfFile
    );

    BOOL NeedReboot;

    if (!UpdateDriverForPlugAndPlayDevices (HWND_DESKTOP, HardwareId, InfFile, 0, &NeedReboot)) {

      TranslateSetupApiError ();

      Error (L"Cannot install driver (%s)", LastErrorText ());

      break;

    }

    RebootRequired = !!NeedReboot;

    Debug (7, L"Device and driver installed, RebootRequired=%u", RebootRequired);

    if (!NoFlush) {

      Debug (7, L"Flushing temporary data");

      FlushStorage ();

    }

    Debug (7, L"Enabling the device");

    if (!EnableDevice (DL, DI, true)) {

      break;

    }

    UINT const PauseAfterEnablingDeviceSec = 3;

    Debug (8, L"Pausing for %u s before checking device status", PauseAfterEnablingDeviceSec);

    Sleep (PauseAfterEnablingDeviceSec * 1000);

    ULONG Status;

    if (GetDeviceStatus (InstanceId, Status, DeviceProblem) && TestBits (Status, DN_HAS_PROBLEM)) {

      Assert (DeviceProblem != 0);

    }

    Success = true;

  } OnlyOnce;

  if (!Success && DeviceInstanceCreated) {

    RemoveDevice (DL, DI, InstanceId);

    DeviceInstanceCreated = false;

  }

  DestroyDL (DL);

  return Success;

}





//
// Entry function
// ==============
//

UINT Main (void) {

  // /opt:ref linker option removes stubs if they are not referenced in object modules.
  // __sse2_available never becomes nonzero so stub functions never called.

  if (__sse2_available != 0) {

    _VEC_memzero ();
    _VEC_memcpy ();

  }

#if _Debug

  GlobalDebugPrintLevel = 6;

#endif

  Debug (8, L"");
  Debug (8, ProductName L" " _US (VersionMajor) L"." _US (VersionMinor));

  UINT Res = 0;

  do {

    VersionInfo.dwOSVersionInfoSize = sizeof (VersionInfo);

    //deprecated, use GetProductInfo?
    Verify (GetVersionEx (&VersionInfo));

    if (VersionInfo.dwPlatformId != VER_PLATFORM_WIN32_NT || VersionInfo.dwMajorVersion < 5) {

      Error (L"OS versions before 5.0 (Windows 2000) are not supported");

      Res = 7;
      break;

    }

    if (!GetTokenInfos ()) {

      Res = 7;
      break;

    }

    PCTSTR CmdLine = GetCommandLine ();

    Debug (7, L"Command line: \"%s\"", CmdLine);  

    CmdLine = GetQuotedString (CmdLine, 0);

    if (!*CmdLine) {

      MessageBox (
        0,
        L"WDM driver manager " _US (VersionMajor) L"." _US (VersionMinor) L"\n"
        L"\n"
        L"Usage: " ProductShortName L" [<options>] <action> <device id> [<INF file>]\n"
        L"\n"
        L"Actions:\n"
        L"\n"
        L"-i - install device  and driver.\n"
        L"-r - remove device and driver.\n"
        L"\n"
        L"<device id> should be a hardware id on installation and a hardware/instance id on removal.\n"
        L"\n"
        L"<INF file> must specify device INF file path (installation only).\n"
        L"\n"
        L"Options:\n"
        L"\n"
        L"-q - no info messages.\n"
        L"-e - no error messages.\n"
        L"-n - don't flush file/registry buffers.\n"
        L"-h <path> - instance id path in the \"Enum\" tree, instead of auto-generated.\n"
        L"",
        L"Information",
        MB_ICONINFORMATION | MB_OK
      );

      Res = 1;
      break;

    }

    bool InstallMode = false;
    SetEOS (InstancePath);

    DWORD OptionMask = 0;

    CmdLine = SkipBlank (CmdLine);

    while (*CmdLine == L'-') {
    
      ++CmdLine;
      
      BYTE const Option = BYTE (*CmdLine++);

      if (!Option) {
      
        Error (L"Missing option letter after the dash");

        Res = 2;
        break;

      }

      if (Option < L'a' || Option > L'z') {

        Error (L"\"-%c\" must be a letter", Option);

        Res = 2;
        break;

      }

      Debug (7, L"Got option: \"%c\"", Option);

      DWORD const Mask = 1u << (Option - L'a');

      if (TestBits (OptionMask, Mask)) {

        Error (L"Option \"-%c\" specified multiple times", Option);

        Res = 2;
        break;

      }

      switch (Option) {

        case L'q': {

          NoInfoMessages = true;

          break;

        }

        case L'e': {

          NoErrorMessages = true;

          break;

        }

        case L'n': {

          NoFlush = true;

          break;

        }

        case L'h': {

          CmdLine = SkipBlank (CmdLine);

          CmdLine = GetQuotedString (CmdLine, InstancePath);

          if (CmdLine != nullptr && !IsEOS (InstancePath)) {

            Debug (8, L"Got instance path: \"%s\"", InstancePath);

          } else {

            Error (L"No instance path");

            Res = 2;

          }

          break;

        }

        case L'i': {

          InstallMode = true;

          break;

        }

        case L'r': {

          InstallMode = false;

          break;

        }

        default: {

          Error (L"Invalid option \"-%c\"", Option);

          Res = 2;
          break;

        }

      } // switch (Option)

      if (Res != 0) {

        break;

      }

      SetBits (OptionMask, Mask);

      CmdLine = SkipBlank (CmdLine);

    } // Option loop

    if (Res != 0) {

      break;

    }

    if (
      !!TestBits (OptionMask, (1 << (L'i' - L'a')))
      == !!TestBits (OptionMask, (1 << (L'r' - L'a')))
    ) {

      Error (L"Invalid -i/-r option combination");

      Res = 2;
      break;

    }

    wchar_t HwId [256];
    wchar_t InfFile [512];

    CmdLine = GetQuotedString (CmdLine, HwId);

    if (!CmdLine) {

      Error (L"Device Id not specified");

      Res = 2;
      break;

    }

    CmdLine = SkipBlank (CmdLine);

    CmdLine = GetQuotedString (CmdLine, InfFile);

    if (!CmdLine) {

      if (!InstallMode) {

        SetEOS (InfFile);

      } else {

        Error (L"INF file not specified");

        Res = 2;
        break;

      }

    }

    ProcessingStarted = true;

    wchar_t FullInfFilePath [MAX_PATH];

    if (InstallMode) {

      DWORD const Len = GetFullPathName (
        InfFile, NumArrayElems (FullInfFilePath), FullInfFilePath, 0
      );

      if (Len == 0) {

        Error (L"Cannot get full path for %s (%s)", InfFile, LastErrorText ());

        Res = 3;
        break;

      }

    }

    if (InstallMode && IsDevicePresent (HwId)) {

      Error (L"Device for %s already installed", HwId);

      Res = 6;
      break;

    }

    bool RebootRequired = false;
    ULONG DeviceProblem = 0;

    bool const Success = InstallMode? (
      Install (HwId, FullInfFilePath, RebootRequired, DeviceProblem)
    ) : (
      Uninstall (HwId, RebootRequired)
    );

    if (!Success) {

      Res = 4;
      break;

    }

    wchar_t Msg [128];

    wsprintf (Msg, L"Device %sinstalled successfully", InstallMode? L"" : L"un");

    UINT MsgType;
    PTSTR p = Msg + wcslen (Msg);

    if (RebootRequired || DeviceProblem != 0) {

      Res = 5;
      MsgType = MB_ICONWARNING;

      wcscpy (p, L" but");
      p += wcslen (p);

      if (DeviceProblem != 0) {

        wsprintf (p, L" has problem %u/%#x", DeviceProblem, DeviceProblem);
        p += wcslen (p);

      }

      if (RebootRequired) {

        if (DeviceProblem != 0) {

          wcscpy (p, L" and");
          p += wcslen (p);

        }

        wcscpy (p, L" reboot is required");
        p += wcslen (p);

      }

    } else {

      Res = 0;
      MsgType = MB_ICONINFORMATION;

    }

    if (MsgType != MB_ICONINFORMATION && !NoErrorMessages || !NoInfoMessages) {

      Message (MsgType | MB_OK, L"%s", Msg);

    }

  } OnlyOnce;

  Debug (8, L"Exiting (%u)", Res);

  return Res;

}





//
// Light-weight entry point if linked with /ENTRY option. Can be removed.
//

//__pragma (message ("##### Debug-only code in effect! #####"))
//__pragma (comment (linker, "/entry:WinAppEntry"))

extern "C" void _cdecl WinAppEntry (void * const Arg) {

  Unused (Arg);

  ExitProcess (Main ());

}





//
// Typical entry point
//

int WINAPI WinMain (HINSTANCE const Instance, HINSTANCE const PrevInstance, PSTR const CmdLine, int const CmdShow) {

  // Make arguments "referenced"

  Unused (Instance && PrevInstance && CmdLine && CmdShow);

  return static_cast <int> (Main ());

}
